# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Qindel Formación y Servicios S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
@author: Natalia Serrano, nserrano at qindel dot com
"""
# pylint: disable=unused-wildcard-import,wildcard-import

import os
import time
import random
import subprocess
import threading

from configparser import NoOptionError
from opengnsys import REST
from opengnsys.log import logger
from .server_worker import ServerWorker

## https://stackoverflow.com/questions/6893968/how-to-get-the-return-value-from-a-thread
class ThreadWithResult (threading.Thread):
    def run (self):
        try:
            self.result = None
            if self._target is not None:
                try:
                    self.result = self._target (*self._args, **self._kwargs)
                except Exception as e:
                    self.result = { 'res': 2, 'der': f'got exception: ({e})' }    ## res=2 as defined in ogAdmClient.c:2048
        finally:
            # Avoid a refcycle if the thread is running a function with an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs

class ogLiveWorker(ServerWorker):
    thread_list = {}

    def interfaceAdmin (self, method, parametros=[]):
        exe = '{}/{}'.format (self.pathinterface, method)
        ## for development only. Will be removed when the referenced bash code (/opt/opengnsys/lib/engine/bin/*.lib) is translated into python
        devel_bash_prefix = '''
            PATH=/opt/opengnsys/scripts/:$PATH;
            for I in /opt/opengnsys/lib/engine/bin/*.lib; do source $I; done;
            for i in $(declare -F |cut -f3 -d" "); do export -f $i; done;
        '''

        if parametros:
            proc = ['bash', '-c', '{} set -x; bash -x {} {}; set +x'.format (devel_bash_prefix, exe, ' '.join (parametros))]
        else:
            proc = ['bash', '-c', '{} set -x; bash -x {}; set +x'.format (devel_bash_prefix, exe)]
        logger.debug ('subprocess.run ("{}", capture_output=True)'.format (proc))
        p = subprocess.run (proc, capture_output=True)
        ## DEBUG
        logger.info (f'stdout follows:')
        for l in p.stdout.strip().decode ('utf-8').splitlines():
            logger.info (f'  {l}')
        logger.info (f'stderr follows:')
        for l in p.stderr.strip().decode ('utf-8').splitlines():
            logger.info (f'  {l}')
        ## /DEBUG
        if 0 != p.returncode:
            cmd_txt = ' '.join (proc)
            logger.error (f'command ({cmd_txt}) failed, stderr follows:')
            for l in p.stderr.strip().decode ('utf-8').splitlines():
                logger.error (f'  {l}')
            raise Exception (f'command ({cmd_txt}) failed, see log for details')
        return p.stdout.strip().decode ('utf-8')

    def tomaIPlocal (self):
        try:
            self.IPlocal = self.interfaceAdmin ('getIpAddress')
        except Exception as e:
            logger.error (e)
            logger.error ('No se ha podido recuperar la dirección IP del cliente')
            return False
        logger.info ('local IP is "{}"'.format (self.IPlocal))
        return True

    def enviaMensajeServidor (self, path, obj={}):
        obj['iph'] = self.IPlocal          ## Ip del ordenador
        obj['ido'] = self.idordenador      ## Identificador del ordenador
        obj['npc'] = self.nombreordenador  ## Nombre del ordenador
        obj['idc'] = self.idcentro         ## Identificador del centro
        obj['ida'] = self.idaula           ## Identificador del aula

        res = self.REST.sendMessage ('/'.join ([self.name, path]), obj)

        if (type (res) is not dict):
            #logger.error ('No se ha podido establecer conexión con el Servidor de Administración')   ## Error de conexión con el servidor
            logger.debug (f'res ({res})')
            logger.error ('Error al enviar trama ***send() fallo')
            return False

        return res

    ## en C, esto envia una trama de respuesta al servidor. Devuelve un boolean
    ## en python, simplemente termina de construir la respuesta y la devuelve; no envía nada por la red. El caller la usa en return() para enviar implícitamente la respuesta
    def respuestaEjecucionComando (self, cmd, herror, ids=None):
        if ids:                 ## Existe seguimiento
            cmd['ids'] = ids    ## Añade identificador de la sesión

        if 0 == herror:  ## el comando terminó con resultado satisfactorio
            cmd['res'] = 1
            cmd['der'] = ''
        else:            ## el comando tuvo algún error
            cmd['res'] = 2
            cmd['der'] = self.tbErroresScripts[herror]           ## XXX

        return cmd

    def cargaPaginaWeb (self, url=None):
        if (not url): url = self.urlMenu
        os.system ('pkill -9 browser')

        p = subprocess.Popen (['/opt/opengnsys/bin/browser', '-qws', url])
        try:
            p.wait (2)       ## if the process dies before 2 seconds...
            logger.error ('Error al ejecutar la llamada a la interface de administración')
            logger.error ('Error en la creación del proceso hijo')
            logger.error ('return code "{}"'.format (p.returncode))
            return False
        except subprocess.TimeoutExpired:
            pass

        return True

    def muestraMenu (self):
        self.cargaPaginaWeb()

    def muestraMensaje (self, idx):
        self.cargaPaginaWeb (f'{self.urlMsg}?idx={idx}')

    def LeeConfiguracion (self):
        try:
            parametroscfg = self.interfaceAdmin ('getConfiguration')   ## Configuración de los Sistemas Operativos del cliente
        except Exception as e:
            logger.error (e)
            logger.error ('No se ha podido recuperar la dirección IP del cliente')
            return None
        logger.debug ('parametroscfg ({})'.format (parametroscfg))
        return parametroscfg

    def onActivation (self):
        if not os.path.exists ('/scripts/oginit'):
            ## no estamos en oglive, este modulo no debe cargarse
            ## esta lógica la saco de src/opengnsys/linux/operations.py, donde hay un if similar
            raise Exception ('Refusing to load within an operating system')

        self.pathinterface   = None
        self.IPlocal         = None     ## Ip del ordenador
        self.idordenador     = None     ## Identificador del ordenador
        self.nombreordenador = None     ## Nombre del ordenador
        self.cache           = None
        self.idproautoexec   = None
        self.idcentro        = None     ## Identificador del centro
        self.idaula          = None     ## Identificador del aula

        try:
            url                = self.service.config.get (self.name, 'remote')
            loglevel           = self.service.config.get (self.name, 'log')
            self.pathinterface = self.service.config.get (self.name, 'pathinterface')
            self.urlMenu       = self.service.config.get (self.name, 'urlMenu')
            self.urlMsg        = self.service.config.get (self.name, 'urlMsg')
        except NoOptionError as e:
            logger.error ("Configuration error: {}".format (e))
            raise e
        logger.setLevel (loglevel)
        self.REST = REST (url)

        if not self.tomaIPlocal():
            raise Exception ('Se han generado errores. No se puede continuar la ejecución de este módulo')

    def _long_running_job (self, name, f, args):
        self._join_threads()

        any_job_running = False
        for k in self.thread_list:
            if self.thread_list[k]['running']:
                any_job_running = True
                break
        if any_job_running:
            logger.info ('some job is already running, refusing to launch another one')
            return { 'job_id': None, 'message': 'some job is already running, refusing to launch another one' }

        job_id = '{}-{}'.format (name, ''.join (random.choice ('0123456789abcdef') for _ in range (8)))
        self.thread_list[job_id] = {
            'thread': ThreadWithResult (target=f, args=args),
            'starttime': time.time(),
            'running': True,
            'result': None
        }
        self.thread_list[job_id]['thread'].start()
        return { 'job_id': job_id }

    def _join_threads (self):
        for k in self.thread_list:
            logger.debug (f'considering thread ({k})')
            elem = self.thread_list[k]
            if 'thread' in elem:
                elem['thread'].join (0.05)
                if not elem['thread'].is_alive():
                    logger.debug (f'is no longer alive, k ({k}) thread ({elem["thread"]})')
                    elem['running'] = False
                    elem['result'] = elem['thread'].result
                    del elem['thread']
